#include "constant.h"

#include "siv_window.h"
#include "sfx_menu.h"

#include <StorageKit.h>
#include <TranslationKit.h>

#include <stdio.h>

#include <GameKit.h>


/*
	this message filter override standard ScrollBar behaviours:
	- when the user moves the knob, B_VALUE_CHANGED messages are sent to the window,
	and the window dispatch it to the scrollbar, which in turn call SetValue->ValueChanged->ScrollTo
	- this filter will directly call a specific ScrollTo depending on SB orientation
*/
class KValueChangedFilter : public BMessageFilter
{
public:

	KValueChangedFilter(KBitmapView *owner):BMessageFilter(B_VALUE_CHANGED)
	{
		fOwner=owner;
	}
	
	virtual filter_result Filter(BMessage *message, BHandler **target)
	{
		BScrollBar *sc=cast_as(*target,BScrollBar);
		if (sc!=NULL&&(sc->Target()==fOwner)) {
			fOwner->ScrollTo(message->FindInt32("value"),sc->Orientation());
			return B_SKIP_MESSAGE;
		}
		return B_DISPATCH_MESSAGE;
	}
	
private:

	KBitmapView *fOwner;
};

KBitmapView::KBitmapView(BRect rect,const char *name,uint32 resizingMode, uint32 flags)
:BView(rect,name,resizingMode,flags|B_WILL_DRAW|B_FRAME_EVENTS)
{
	fBitmap=NULL;
	fDblBitmap=NULL;
	fScrollRect=BRect(0,0,0,0);
	SetViewColor(B_TRANSPARENT_32_BIT);
	fZoomFactor=1.;
	fActualBounds=Bounds();
	fLockScrollTo=false;
}

KBitmapView::~KBitmapView()
{
	clean_up();
}

void KBitmapView::clean_up()
{
	delete fBitmap;
	delete fDblBitmap;
	fBitmap=NULL;
	fDblBitmap=NULL;
}
	
void KBitmapView::SetZoomFactor(float factor)
{
	fZoomFactor=factor;
	compute_actual_bounds();
	ScrollTo(BPoint(0,0));
}

float KBitmapView::ZoomFactor()
{
	return fZoomFactor;
}

void KBitmapView::compute_actual_bounds()
{
	if (fBitmap!=NULL) {
		BPoint origin=fActualBounds.LeftTop();
		fActualBounds=fBitmap->Bounds();
		fActualBounds.left/=fZoomFactor;
		fActualBounds.right/=fZoomFactor;
		fActualBounds.top/=fZoomFactor;
		fActualBounds.bottom/=fZoomFactor;
		fActualBounds.OffsetTo(origin);
	}
}

void KBitmapView::resize_to_fit()
{
	if (fBitmap!=NULL) {
		float scaling=1;
		BScreen screen(Window());
		BRect fMaxBounds=screen.Frame();
		fMaxBounds.OffsetTo(0,0);
		fMaxBounds.right-=5;
		fMaxBounds.bottom-=30;
		scaling=fBitmap->Bounds().Width()/fMaxBounds.Width();
		SetZoomFactor(scaling);
		if (fActualBounds.Height()>fMaxBounds.Height()) {
			scaling=fBitmap->Bounds().Height()/fMaxBounds.Height();
			SetZoomFactor(scaling);
		}
	}
}
	
void KBitmapView::Draw(BRect update)
{
	if (fBitmap!=NULL) {
		BRect x=update;
		x.OffsetBy(fActualBounds.LeftTop());
		x.left*=fZoomFactor;
		x.right*=fZoomFactor;
		x.top*=fZoomFactor;
		x.bottom*=fZoomFactor;
		DrawBitmap(fBitmap,x,update);
		//DrawBitmap(fBitmap,update,update);
	} else {
		FillRect(update);
	}
	/*
	BPoint where;
	uint32 buttons;
	GetMouse(&where,&buttons,false);
	SetPenSize(4);
	SetHighColor(0,100,200);
	SetDrawingMode(B_OP_BLEND);
	StrokeRect(BRect(where.x-100,where.y-100,where.x+100,where.y+100));
	*/
}
	
void KBitmapView::SetTo(BBitmap *bitmap)
{
	clean_up();
	fBitmap=bitmap;
	//compute_actual_bounds();
	//resize_to_fit();
	compute_actual_bounds();
	// ---
	ScrollTo(BPoint(0,0));
	//Invalidate();
	Draw(Bounds());
}

void KBitmapView::AttachedToWindow(void) 
{
	Window()->AddCommonFilter(new KValueChangedFilter(this));
	inherited::AttachedToWindow();
}

void KBitmapView::FrameResized(float width, float height)
{
	if (fBitmap==NULL) return;
	// ---
	BScrollBar *sb;
	if ((sb=ScrollBar(B_HORIZONTAL))!=NULL) {
		sb->SetRange(0,fScrollRect.right=fActualBounds.Width()-Bounds().Width());
		sb->SetProportion(Bounds().Width()/fActualBounds.Width());
	}
	if ((sb=ScrollBar(B_VERTICAL))!=NULL) {
		sb->SetRange(0,fScrollRect.bottom=fActualBounds.Height()-Bounds().Height());
		sb->SetProportion(Bounds().Height()/fActualBounds.Height());
	}
}

void KBitmapView::ScrollTo(int32 value,orientation posture)
{
	if (posture==B_HORIZONTAL)
		ScrollTo(BPoint(value,fActualBounds.top));
	else if (posture==B_VERTICAL)
		ScrollTo(BPoint(fActualBounds.left,value));
}

void KBitmapView::ScrollTo(BPoint origin)
{
	if (fLockScrollTo) return;
	if (origin==fActualBounds.LeftTop()) return;
	fActualBounds.OffsetTo(origin);
	// update the scroll-bars ----
	// Have to lock the ScrollTo, cause SB will call it back with wrong values...
	BScrollBar *sc=NULL;
	fLockScrollTo=true;
	if ((sc=ScrollBar(B_VERTICAL))!=NULL) {
		sc->SetValue(origin.y);
	}
	if ((sc=ScrollBar(B_HORIZONTAL))!=NULL) {
		sc->SetValue(origin.x);
	}
	fLockScrollTo=false;
	// invalidate everything
	Invalidate();
}

BBitmap *KBitmapView::ReleaseBitmap()
{
	BBitmap *returned=fBitmap;
	fBitmap=NULL;
	return returned;
}
	
void KBitmapView::GetPreferredSize(float *width, float *height)
{
	if (fBitmap!=NULL) {
		*width=fActualBounds.Width();
		*height=fActualBounds.Height();
	} else {
		puts("null");
		*width=Bounds().Width();
		*height=Bounds().Height();
	}
}

void KBitmapView::MouseDown(BPoint where)
{
	if (modifiers()&B_OPTION_KEY) {
		fMouseOffset=where;
		fScrolling=true;
		SetMouseEventMask(B_POINTER_EVENTS,B_NO_POINTER_HISTORY);
	} else {
	BPoint pos;
	uint32 buttons=1,mdf=modifiers();
	bool lock_modifiers=false;
	float zoom_speed=.996;
	if (mdf&B_CONTROL_KEY) {
		zoom_speed=1/zoom_speed;
		lock_modifiers=true;
	} else if (mdf!=0) {
		lock_modifiers=true;
	}
	BPoint origin=where,delta=BPoint(0,0);
	while (buttons!=0) {
		GetMouse(&pos,&buttons,true);
		///
		fZoomFactor*=zoom_speed;
		compute_actual_bounds();
		FrameResized(Bounds().Width(),Bounds().Height());
		if (modifiers()&B_OPTION_KEY) {
			delta=pos-origin;
		} else {
			delta=BPoint(0,0);
		}
		origin=fActualBounds.LeftTop()+pos-delta;
		origin.x/=zoom_speed;
		origin.y/=zoom_speed;
		origin-=pos;
		ScrollTo(origin);
		origin=pos;
		//Draw(Bounds());
		//puts("scroolto");
		//Invalidate();
		// ---
		mdf=modifiers();
		if (!lock_modifiers&&mdf&B_SHIFT_KEY) {
			zoom_speed*=.9999;
		} else if (!lock_modifiers&&mdf&B_CONTROL_KEY) {
			zoom_speed*=1.0001;
		} else if (mdf==0) {
			lock_modifiers=false;
		}
		snooze(10000);
	}
	}
}

void KBitmapView::MouseUp(BPoint where)
{
	fScrolling=false;
}

void KBitmapView::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
	if (fScrolling) {
		BPoint offset=Bounds().LeftTop()+fMouseOffset-point;
		//fMouseOffset=point;
		offset.ConstrainTo(fScrollRect);
		ScrollTo(BPoint(offset.x,offset.y));
	}
}

// SIVWindow <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

SIVWindow::SIVWindow(BRect frame,const char *title)
:BWindow(frame,title,B_DOCUMENT_WINDOW,B_NOT_RESIZABLE|B_ASYNCHRONOUS_CONTROLS)
{
	Init();
}

void SIVWindow::Init()
{
	fMainView=new KBitmapView(Bounds(),"main_view",B_FOLLOW_ALL,0);
	fScrollView=new BScrollView("scroll_view",fMainView,B_FOLLOW_ALL,0,true,true,B_NO_BORDER);
	AddChild(fScrollView);
	fMainView->AddFilter(new BMessageFilter(B_MOUSE_DOWN,filter));
	//
	//(fScrollView->ScrollBar(B_VERTICAL))->AddFilter(new BMessageFilter(B_ANY_DELIVERY,B_ANY_SOURCE,spy_filter));
	//be_app->AddFilter(new BMessageFilter(B_ANY_DELIVERY,B_ANY_SOURCE,spy_filter));
	//
	fScrollView->MoveTo(0,0);
	fScrollView->ResizeTo(Bounds().Width(),Bounds().Height());
	//
	BMessage *invoke;
	AddShortcut('N',0,invoke=new BMessage(K_ACTION));
	invoke->AddInt32("action",K_ACTION_GO_FORWARD);
	AddShortcut('R',0,invoke=new BMessage(K_UPDATE));
}
	
/* load an image using translation kit, and display it in the main view */
status_t SIVWindow::LoadImage(const char *path)
{
	// be sure the window is locked before making changes
	if (Lock()) {
		// first, try to load the image
		BBitmap *image=BTranslationUtils::GetBitmapFile(path);
		if (image==NULL) return B_ERROR;
		fImagePath.SetTo(path);
		// loading ok, display it
		fMainView->SetTo(image);
		// resize the window
		UpdateDisplay();
		//PostMessage(K_UPDATE,this);
		Unlock();
		return B_OK;
	} else
		return B_ERROR;
}

void SIVWindow::UpdateDisplay()
{
	BScreen screen(this);
	float width,height;
	fMainView->GetPreferredSize(&width,&height);
	BRect frame=Frame();
	// resize in order that fMainview got the correct size
	frame.right=frame.left+Bounds().Width()+width-fMainView->Bounds().Width(),
	frame.bottom=frame.top+Bounds().Height()+height-fMainView->Bounds().Height();
	// take care of screen size
	SetFlags(B_NOT_RESIZABLE);
	if (frame.bottom>screen.Frame().bottom) frame.OffsetTo(frame.left,24);
	if (frame.right>screen.Frame().right) frame.OffsetTo(1,frame.top);
	if (frame.bottom>screen.Frame().bottom) {
		frame.bottom=screen.Frame().bottom-30;
		SetFlags(0);
	}
	if (frame.right>screen.Frame().right) {
		frame.right=screen.Frame().right-5;
		SetFlags(0);
	}
	// set the limit of the window
	float minWidth,maxWidth,minHeight,maxHeight;
	GetSizeLimits(&minWidth,&maxWidth,&minHeight,&maxHeight);
	maxWidth=frame.Width();maxHeight=frame.Height();
	SetSizeLimits(minWidth,maxWidth,minHeight,maxHeight);
	SetZoomLimits(maxWidth,maxHeight);
	// resize the window (nb: doing the resizing before setting the limits seems not to work properly...the window is not resize ???)
	MoveTo(frame.left,frame.top);
	ResizeTo(frame.Width(),frame.Height());
	// force the window to become frontmost
	SetWorkspaces(B_CURRENT_WORKSPACE);// if the window is already open in another workspace
	// there's no SendFront method...
	if (!IsFront()) {
		//Hide();
		Show();
	}
}

void SIVWindow::ForceMouseWithin()
{
	BPoint where;
	uint32 buttons;
	fMainView->GetMouse(&where,&buttons,false);
	where=fMainView->ConvertToScreen(where);
	if (!Frame().Contains(where)) {
		BRect frame=Frame();
		frame.InsetBy(5,5);
		where.ConstrainTo(frame);
		set_mouse_position(where.x,where.y);
	}
}


void SIVWindow::DispatchMessage(BMessage *message, BHandler *target)
{
	switch (message->what) {
		case B_VALUE_CHANGED:{
			message->PrintToStream();
			break;
		}
		default:inherited::DispatchMessage(message,target);
	}
}
	
void SIVWindow::MessageReceived(BMessage *message)
{
	switch (message->what) {
		case B_SIMPLE_DATA:{
			entry_ref ref;
			if (message->FindRef("refs",&ref)==B_OK) {
				BEntry entry(&ref);
				BPath path(&entry);
				LoadImage(path.Path());
			}
			break;
		}
		case B_SET_PROPERTY:{
			int32 index,what;
			BMessage specifier;
			char *property=NULL;
			message->GetCurrentSpecifier(&index,&specifier,&what,&property);
			if ((strcmp(property,"ZoomFactor") == 0)&&(what == B_DIRECT_SPECIFIER)) {
				float factor=message->FindFloat("data");
				fMainView->SetZoomFactor(factor);
				UpdateDisplay();
				fMainView->Invalidate();
			}
			break;
		}
		case K_UPDATE:
			UpdateDisplay();
		case K_ACTION:{
			int32 action=message->FindInt32("action");
			switch (action) {
				case K_ACTION_GO_FORWARD:{
					// check if it's another action pending...
					BMessage *a_mess=NULL;
					int32 mess_index=0;
					while ((a_mess=MessageQueue()->FindMessage(K_ACTION,mess_index++))!=NULL) {
						if (a_mess->FindInt32("action")==K_ACTION_GO_FORWARD) {
							return;
						}
					}
					BEntry entry(fImagePath.Path());
					if (entry.InitCheck()==B_OK) {
						BDirectory dir;
						entry.GetParent(&dir);
						int32 count=dir.CountEntries();
						entry_ref ref,xref;
						entry.GetRef(&ref);
						dir.Rewind();
						
						bool grab=false;
						status_t err=B_OK;
						
						while (count>0) {
						
							err=dir.GetNextRef(&xref);
							
							if (err!=B_OK) {
								dir.Rewind();
							} else if (grab) {
								count--;
								entry.SetTo(&xref);
								BPath path(&entry);
								if (LoadImage(path.Path())==B_OK) {
									ForceMouseWithin();
									snooze(100000);
									break;
								}
							} else {
								grab=(ref==xref);
							}
						}
					}
					break;
				}// end K_ACTION_GO_FORWARD
			}// end action switch
			break;
		}// end K_ACTION
		default:
			inherited::MessageReceived(message);
	}
}

BHandler *SIVWindow::ResolveSpecifier(BMessage *message, int32 index, BMessage *specifier, int32 what, const char *property)
{
	if ((strcmp(property,"ZoomFactor") == 0)&&(what == B_DIRECT_SPECIFIER)) {
		return this;
	} else 
		return inherited::ResolveSpecifier(message,index,specifier,what,property);
}

void SIVWindow::FrameMoved(BPoint screenPoint)
{
	//if (screenPoint.y<20) MoveTo(screenPoint.x,20);
	//if (screenPoint.x<0) MoveTo(0,screenPoint.y);
}

bool SIVWindow::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

void SIVWindow::InvokePopUp(BPoint where)
{
	SFXPopUpMenu *popup=new SFXPopUpMenu("xxx");
	BMenuItem *item;
	BMessage *invoke;
	popup->AddItem(item=new BMenuItem("Next image",invoke=new BMessage(K_ACTION),'N'));
	invoke->AddInt32("action",K_ACTION_GO_FORWARD);
	item->SetTarget(this);
	popup->AddItem(new BMenuItem("blabla",new BMessage('xxxx')));
	popup->AddItem(new BMenuItem("blabla",new BMessage('xxxx')));
	popup->AddItem(new BMenuItem("blabla",new BMessage('xxxx')));
	popup->AddSeparatorItem();
	// zoom factor
	invoke=new BMessage(B_SET_PROPERTY);
	invoke->AddSpecifier("ZoomFactor");
	invoke->AddFloat("data",1.);
	popup->AddItem(item=new BMenuItem("zoom 100%",invoke));
	item->SetTarget(this);
	invoke=new BMessage(invoke);
	invoke->ReplaceFloat("data",2.);
	popup->AddItem(item=new BMenuItem("zoom 50%",invoke));
	item->SetTarget(this);
	invoke=new BMessage(invoke);
	invoke->ReplaceFloat("data",.5);
	popup->AddItem(item=new BMenuItem("zoom 200%",invoke));
	item->SetTarget(this);
	// display mode
	//invoke=new BMessage(B_SET_PROPERTY);
	//invoke->AddSpecifier("DisplayMode");
	//
	popup->AddSeparatorItem();
	popup->AddItem(item=new BMenuItem("About SIV",new BMessage(B_ABOUT_REQUESTED)));
	item->SetTarget(be_app);
	// ---
	popup->Go(ConvertToScreen(where),true,false,false);
	delete popup;
}

static filter_result SIVWindow::filter
		(BMessage *message, BHandler **target,BMessageFilter *messageFilter)
{
	BView *view=cast_as(*target,BView);
	SIVWindow *myself=cast_as(messageFilter->Looper(),SIVWindow);
	if (view!=NULL&&myself!=NULL) {
		BPoint where;
		message->FindPoint("where",&where);
		where=view->ConvertToParent(where);
		int32 buttons = myself->CurrentMessage()->FindInt32("buttons"); 
		if (buttons==B_SECONDARY_MOUSE_BUTTON) {
			myself->InvokePopUp(where);
			return B_SKIP_MESSAGE;
		}
	}
	return B_DISPATCH_MESSAGE;
}

static filter_result SIVWindow::spy_filter
		(BMessage *message, BHandler **target,BMessageFilter *messageFilter)
{
	message->PrintToStream();
	return B_DISPATCH_MESSAGE;
}
	
	
	